#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>
#include <time.h>
#include <sys/time.h>
#include "sha1.h"

#define FALSE (0)
#define TRUE (!FALSE)

#define VectorType uint64_t
#define VectorPRIx PRIx64

#define LOG_INTERVAL (2L << 22)
#define REFRESH_INTERVAL (1L << 15)

// must be 2*N
#define KEYLEN (16)
#define KEYUNIT (2)

#define B64UNIT (6)

typedef struct {
	uint32_t x, y, z, w, temp;
} rand_params;

static rand_params randp;

#define rand_init(params, seed) \
	((params).x=(seed)/*123456789UL*/, (params).y=362436069UL, \
	(params).z=521288629UL, (params).w=88675123UL)

#define rand_get(params) \
	((params).temp=(params).x^((params).x<<11), \
	(params).x=(params).y, (params).y=(params).z, (params).z=(params).w, \
	(params).w^=((params).w>>19)^(params).temp^((params).temp>>8))

#define RAND(n) ((int) (rand_get(randp)%(n)))

#define SWAP32(x) (((x) >> 24) | (((x)&0x00FF0000) >> 8) | (((x)&0x0000FF00) << 8) | (x << 24))

typedef union {
	VectorType vector[SHA1_HASH_SIZE*sizeof(VectorType)/sizeof(uint32_t)];
	uint32_t H[SHA1_HASH_SIZE];
	uint8_t  sha1[SHA1_HASH_SIZE*sizeof(uint32_t)];
} Sha1Vector;

static uint8_t base64Table[] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	"abcdefghijklmnopqrstuvwxyz"
	"0123456789./";

void printResult();
void makeMessage();
void calc_sha1();
void printSearchVector(Sha1Vector vector, Sha1Vector mask);
void makeSearchVector(char *b64prefix);
void changeKey(uint8_t *target);
void refreshKey(uint8_t *target);
void makeKey(uint8_t *target);
static void initRandVector();
static int isValidVector(int i, int j);

static uint8_t randVector[0xffff][2];
static int  randVectorSize = 0;

#define SEARCH_VECTOR_MAX (256)
static Sha1Vector searchVector[SEARCH_VECTOR_MAX];
static Sha1Vector searchVectorMask[SEARCH_VECTOR_MAX];
static int        searchVectorLen = 0;

static union {
	uint32_t message[SHA1_STEP_SIZE];
	uint8_t key[KEYLEN+1];
} M;
static Sha1Vector H;

int
main(int argc, char *argv[]) {
	if (argc < 2) {
		fprintf(stderr, "usage: %s searchPrefix..\n", argv[0]);
		return 1;
	}
	for (int i = 1; i < argc; i++) {
		makeSearchVector(argv[i]);
		printSearchVector(searchVector[i-1], searchVectorMask[i-1]);
		//for (int j = 0; j < sizeof(searchVector[0].vector)/sizeof(searchVector[0].vector[0]); j++) {
		//	fprintf(stderr, "%016"VectorPRIx" %016"VectorPRIx"\n", searchVector[i-1].vector[j], searchVectorMask[i-1].vector[j]);
		//}
	}
	
	struct timeval tv;
	gettimeofday(&tv, 0);
	uint32_t seed = ((uint32_t)tv.tv_sec) ^ ((uint32_t)tv.tv_usec);

	rand_init(randp, seed);
	initRandVector();
	makeMessage();
	uint64_t nkeys  = 0;
	uint64_t nfound = 0;
	struct timeval tv_start, tv_end;
	gettimeofday(&tv_start, 0);
	while(TRUE) {
		calc_sha1();
		for (int i = 0; i < searchVectorLen; i++) {
			if ((H.vector[0] & searchVectorMask[i].vector[0]) == searchVector[i].vector[0]) {
				// ĤΥХȤå롣
				int matched = TRUE;
				for (int j = 1; j < sizeof(H.vector)/sizeof(H.vector[0]); j++) {
					if ((H.vector[j] & searchVectorMask[i].vector[j]) != searchVector[i].vector[j]) {
						matched = FALSE;
						break;
					}
				}
				if (matched) {
					// 
					nfound++;
					printResult();
					break;
				}
			}
		}
		
		nkeys++;
		if (nkeys % LOG_INTERVAL == 0) {
			gettimeofday(&tv_end, 0);
			double delta = (double)(tv_end.tv_sec - tv_start.tv_sec)
						+ (double)(tv_end.tv_usec - tv_start.tv_usec) / 1000000.0;
			fprintf(stderr, "%.2f Mtrips/s (%ld keys in %f sec) %"PRId64"/%"PRId64"\n",
				LOG_INTERVAL * 1.0 / delta / 1024 / 1024,
				LOG_INTERVAL, delta, nfound, nkeys);
			gettimeofday(&tv_start, 0);
		}
		// ѹ
		if (nkeys % REFRESH_INTERVAL == 0) {
			// եå
			refreshKey(M.key);
		} else {
			changeKey(M.key);
		}
	}
}

void
printB64(FILE* out, uint8_t *str, uint32_t len) {
	int printLen = 0;
	for (int b64offset = 0; b64offset < len; b64offset += 3) {
		uint32_t bitmap = 0;
		bitmap |= str[b64offset] << 8*2;
		if (b64offset+1 < len)
			bitmap |= str[b64offset+1] << 8*1;
		if (b64offset+2 < len)
			bitmap |= str[b64offset+2];
		for (int b64offset2 = 3; b64offset2 >= 0; b64offset2--) {
			uint32_t unitMask = (1 << B64UNIT)-1;
			uint32_t b64char = ((bitmap & (unitMask << (b64offset2 * B64UNIT))) >> (b64offset2 * B64UNIT));
			fprintf(out, "%c", base64Table[b64char]);
			printLen++;
			if (12 == printLen)
				fprintf(out, " ");
		}
	}
}

void
printResult() {
	char key[KEYLEN+1];
	memset(key, 0, sizeof(key));
	memcpy(key, M.key, KEYLEN);
	printf("%s ", key);

	// calc_sha1ǤȽŤΤǤǡ
	for (int i = 0; i < 5; i++) {
		H.H[i] = SWAP32(H.H[i]);
	}
	printB64(stdout, H.sha1, SHA1_HASH_SIZE * sizeof(uint32_t));

	//for(int i = 0; i < 4*5; i++) {
	//	printf("%02x", (uint8_t)H.sha1[i]);
	//}
	printf("\n");
	fflush(stdout);
}

void
makeMessage() {
	memset(M.message, 0, sizeof(M.message));
	makeKey(M.key);
	M.key[KEYLEN] = '\x80';
	M.message[SHA1_STEP_SIZE-1] = SWAP32(KEYLEN * 8);
}

void
calc_sha1() {
	// Ҥ
	//memset(message, 0, sizeof(message));
	//memcpy(message, key, KEYLEN);
	//((char *)message)[KEYLEN] = '\x80';
	//message[SHA1_STEP_SIZE-1] = SWAP32(KEYLEN * 8);

	// SHA1
	H.H[0] = 0x67452301;
	H.H[1] = 0xefcdab89;
	H.H[2] = 0x98badcfe;
	H.H[3] = 0x10325476;
	H.H[4] = 0xc3d2e1f0;
	sha1_step(H.H, M.message, 1);
	// makeSearchVector&printResultǤ
	//for (int i = 0; i < 5; i++) {
	//	H.H[i] = SWAP32(H.H[i]);
	//}
}

void
printSearchVector(Sha1Vector vector, Sha1Vector mask) {
	Sha1Vector X;
	for (int i = 0; i < sizeof(X.sha1); i++) {
		X.sha1[i] = '\0';
	}
	for (int i = 0; i < sizeof(X.vector)/sizeof(X.vector[0]); i++) {
		X.vector[i] = vector.vector[i];
	}
	for (int i = 0; i < 5; i++) {
		X.H[i] = SWAP32(X.H[i]);
	}
	fprintf(stderr, "Vector: ");
	printB64(stderr, X.sha1, SHA1_HASH_SIZE * sizeof(uint32_t));
	fprintf(stderr, "\n");

	for (int i = 0; i < sizeof(X.sha1); i++) {
		X.sha1[i] = '\0';
	}
	for (int i = 0; i < sizeof(X.vector)/sizeof(X.vector[0]); i++) {
		X.vector[i] = mask.vector[i];
	}
	for (int i = 0; i < 5; i++) {
		X.H[i] = SWAP32(X.H[i]);
	}
	fprintf(stderr, "Mask  : ");
	printB64(stderr, X.sha1, SHA1_HASH_SIZE * sizeof(uint32_t));
	
	fprintf(stderr, "\n");
}

void
makeSearchVector(char *b64prefix) {
	Sha1Vector vector;
	Sha1Vector vectorMask;
	for (int i = 0; i < sizeof(vector.vector)/sizeof(vector.vector[0]); i++) {
		vector.vector[i]     = 0;
		vectorMask.vector[i] = 0;
	}
	VectorType unitMask   = (1 << B64UNIT) - 1;
	for (int i = 0; b64prefix[i]; i++) {
		char *charPos = strchr((char *)base64Table, b64prefix[i]);
		if (NULL == charPos)
			continue;
		VectorType rawBit = charPos - (char *)base64Table;

		int totalOffset = i * B64UNIT;
		for (int j = 0; j < sizeof(vector.vector)/sizeof(vector.vector[0]); j++) {
			int offset = (j+1) * sizeof(VectorType) * 8 - B64UNIT - totalOffset;
			if (-B64UNIT < offset && offset <= (int)sizeof(VectorType)*8) {
				if (offset < 0) {
					vector.vector[j]     += rawBit   >> -offset;
					vectorMask.vector[j] += unitMask >> -offset;
				} else {
					vector.vector[j]     += rawBit   << offset;
					vectorMask.vector[j] += unitMask << offset;
				}
			}
		}

	}

	for (int i = 0; i < sizeof(vector.vector)/sizeof(vector.vector[0]); i++) {
		searchVector[searchVectorLen].vector[i]     = 0;
		searchVectorMask[searchVectorLen].vector[i] = 0;
	}
	for (int i = 0; i < sizeof(vector.sha1)/sizeof(vector.sha1[0]); i++) {
		VectorType charMask = 0xff;
		int vectorIndex = i / sizeof(VectorType);
		int offset = (sizeof(VectorType)-1) - i % sizeof(VectorType);
		searchVector[searchVectorLen].sha1[i]     = (vector.vector[vectorIndex]     & (charMask << (offset*8))) >> (offset*8);
		searchVectorMask[searchVectorLen].sha1[i] = (vectorMask.vector[vectorIndex] & (charMask << (offset*8))) >> (offset*8);
	}
	for (int i = 0; i < sizeof(vector.H)/sizeof(vector.H[0]); i++) {
		searchVector[searchVectorLen].H[i]     = SWAP32(searchVector[searchVectorLen].H[i]);
		searchVectorMask[searchVectorLen].H[i] = SWAP32(searchVectorMask[searchVectorLen].H[i]);
	}

	searchVectorLen++;
}

void
changeKey(uint8_t *target) {
	int r = RAND(randVectorSize*(KEYLEN/KEYUNIT));
	int nthUnit   = r%(KEYLEN/KEYUNIT);
	int nthVector = r/(KEYLEN/KEYUNIT);
	memcpy(&target[nthUnit*KEYUNIT], randVector[nthVector], KEYUNIT);
}

void
refreshKey(uint8_t *target) {
	int writePos = 0;
	for (writePos = 0; writePos < KEYLEN; writePos += KEYUNIT) {
		int nthVector = RAND(randVectorSize);
		memcpy(&target[writePos], randVector[nthVector], KEYUNIT);
	}
}

void
makeKey(uint8_t *target) {
	refreshKey(target);
	target[KEYLEN] = '\0';
}

void
initRandVector() {
	for (int i = 0x00; i < 0xff; i++) {
		for (int j = 0x00; j < 0xff; j++) {
			if (isValidVector(i, j)) {
				randVector[randVectorSize][0] = i;
				randVector[randVectorSize][1] = j;
				randVectorSize++;
			}
		}
	}
}

int
isValidVector(int i, int j) {
	// Shift_JISȤХ򤹤롣
	// ͻ: http://charset.7jp.net/sjis.html
	
	// 1Хܡ2Х Ⱦʸξ
	if (0x20 <= i && i <= 0x7e) {
		// ASCII
		if (0x20 <= j && j <= 0x7e) {
			// ASCII
			return TRUE;
		}
		if (0xa1 <= j && j <= 0xdf) {
			// Ⱦѥ
			return TRUE;
		}
	}
	if (0xa1 <= i && i <= 0xdf) {
		// Ⱦѥ
		if (0x20 <= j && j <= 0x7e) {
			// ASCII
			return TRUE;
		}
		if (0xa1 <= j && j <= 0xdf) {
			// Ⱦѥ
			return TRUE;
		}
	}
	
	// ʸξ
	if (0x7f != j) { // 2ʸܤ0x7f
		if (0x81 == i) {
			if (0x40 <= j && j <= 0xac) {
				// 
				// ¡áġšơǡȡɡʡˡ̡͡ΡϡСѡҡӡԡա
				// סء١ڡۡܡݡޡߡ
				// 
				return TRUE;
			}
			if (0xb8 <= j && j <= 0xbf) {
				// 
				return TRUE;
			}
			if (0xc8 <= j && j <= 0xce) {
				// ʢˢ̢͢΢Ϣ
				return TRUE;
			}
			if (0xda <= j && j <= 0xe8) {
				// ܢݢޢߢ
				return TRUE;
			}
			if (0xf0 <= j && j <= 0xf7) {
				// 
				return TRUE;
			}
			if (0xfc == j) {
				// 
				return TRUE;
			}
		}
		if (0x82 == i) {
			if (0x4f <= j && j <= 0x58) {
				// 
				return TRUE;
			}
			if (0x60 <= j && j <= 0x79) {
				// £ãģţƣǣȣɣʣˣ̣ͣΣϣУѣңӣԣգ֣ףأ٣
				return TRUE;
			}
			if (0x81 <= j && j <= 0x9a) {
				// 
				return TRUE;
			}
			if (0x9f <= j && j <= 0xf1) {
				// 
				// ¤äĤŤƤǤȤɤʤˤ̤ͤΤϤФѤҤӤ
				// դ֤פؤ٤ڤۤܤݤޤߤ
				// 
				return TRUE;
			}
		}
		if (0x83 == i) {
			if (0x40 <= j && j <= 0x96) {
				// 
				// ¥åĥťƥǥȥɥʥ˥̥ͥΥϥХѥҥӥ
				// ե֥ץإ٥ڥۥܥݥޥߥ
				// 
				return TRUE;
			}
			if (0x9f <= j && j <= 0xb6) {
				// 
				return TRUE;
			}
			if (0xbf <= j && j <= 0xd6) {
				// ¦æĦŦƦǦȦɦʦ˦̦ͦΦϦЦѦҦӦԦզ֦צ
				return TRUE;
			}
		}
		if (0x84 == i) {
			if (0x40 <= j && j <= 0x60) {
				// 
				return TRUE;
			}
			if (0x70 <= j && j <= 0x91) {
				// ѧҧӧԧէ֧קا٧ڧۧܧݧާߧ
				return TRUE;
			}
			if (0x9f <= j && j <= 0xbe) {
				// 
				return TRUE;
			}
		}
		// 
		if (0x88 == i && 0x9f <= j && j <= 0xfc) {
			return TRUE;
		}
		if (0x89 <= i && i <= 0x9f) {
			if (0x98 == i && 0x73 <= j && j <= 0x9e) {
				return FALSE;
			}
			if (0x40 <= j && j <= 0xfc) {
				return TRUE;
			}
		}
		if (0xe0 <= i && i <= 0xe9) {
			if (0x40 <= j && j <= 0xfc) {
				return TRUE;
			}
		}
		if (0xea == i && 0x40 <= j && j <= 0xa4) {
			return TRUE;
		}
	}
	return FALSE;
}
